import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;

/**
 * Dunia tetris!
 * 
 * @author Keenan M. Gebze 
 * @version 1.0
 */
public class TetrisWorld  extends World
{
    Tetromino currentTetromino;
    Set<Tetromino> bag;
    Iterator<Tetromino> iBag;

    Counter scoreCounter = new Counter("Your Score: ");
    Counter bagCounter = new Counter("Tetris Bag: ");
    GreenfootSound bgm;
    
    boolean gameLose= false;
    int bagCount;
    int speed=45;

    /**
     * Konstruktor Dunia Tetris.
     */
    public TetrisWorld()
    {    
        //ukuran standard permainan tetris :D
        super(10, 20, 25, false);
        bgm = new GreenfootSound("tem.mid");
        bag = newBag();
        bagCounter.add(1);
        Greenfoot.setSpeed(speed);
        addObject(scoreCounter, 3, 0);
        addObject(bagCounter, 3, 1);
    }

    public void started() {
        bgm.playLoop();
    }

    /**
     * Disini kita akan membuat gameplaynya!
     */
    public void act() {
        if(((currentTetromino == null) || currentTetromino.isInactive()) && !gameLose) {

            if(bag.isEmpty()) {
                bag = newBag();
                bagCounter.add(1);
                bagCount++;
                if(bagCount >= 3) {
                    Greenfoot.setSpeed(speed + 2);
                    bagCount = 0;
                }
            }
            iBag = bag.iterator();            
            currentTetromino = iBag.next();
            bag.remove(currentTetromino);
            addObject(currentTetromino, getWidth()/2, 1);
            scan();
        }
    }

    public void gameLose() {
        gameLose = true;
        removeObjects(getObjects(Block.class));
        removeObjects(getObjects(Tetromino.class));
        
        //ScoreUploader so = new ScoreUploader("http://games.1free.ws/scripts/tetrisscore.php");
        //String nama = ScoreUploader.promptName();
        //String message = "username="+nama+"&"+"userscore=" + scoreCounter.getValue();
        //so.send(message);
        
    }

    private Set<Tetromino> newBag() {
        bag = new HashSet(7);        
        bag.add(new IBlock());
        bag.add(new TBlock());
        bag.add(new OBlock());
        bag.add(new LBlock());
        bag.add(new JBlock());
        bag.add(new SBlock());
        bag.add(new ZBlock());
        return bag;
    }

    public void scan() {
        int multiplier = 0;
        for(int y=0; y<getHeight(); y++ ){
            if(isLineFull(y)) {
                deleteLine(y);
                multiplier++;
            }
        }
        scoreCounter.add(speed*(multiplier*2));
    }

    public boolean isLineFull(int y) {
        for(int x=0; x<getWidth(); x++) {
            List li = getObjectsAt(x, y, Block.class);
            Iterator io = li.iterator();
            if(io.hasNext()) 
            {
                Block a = (Block) io.next();
            }else {
                return false;
            }
        }
        return true;
    }

    public void deleteLine(int y) {
        for(int x=0; x<getWidth(); x++) {
            removeObjects(getObjectsAt(x, y, Block.class));
        }
        List objects = getObjects(Block.class);
        Iterator oi = objects.iterator();
        while(oi.hasNext()) {
            Actor moveDown = (Actor) oi.next();
            if(moveDown.getY() < y) {
                moveDown.setLocation(moveDown.getX(), moveDown.getY()+1);
            }
        }
        play(80, 300);
        play(83, 300);
        play(85, 300);
        play(88, 300);
    }

    public void play(int pitch, int length)
    {
        //player.play(pitch, length);
    }

    public void stopped() {
        //if(player != null) {
        //    player.turnAllOff();
        //}
    }

}
