import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
//import java.awt.Color;

/**
 * Bongkahan tetris. Di kelas orang tua ini kita membuat sifat dasar
 * dari bongkahan-bongkahan tetris.
 * Sebenarnya kelas ini dan anak kelas ini bukan merupakan 'bongkahannya',
 * tetapi sesuatu yang tidak terlihat dan mengatur blok-blok individu.
 * 
 * @author Keenan Mandela Gebze.
 * @version 1.0
 */
public abstract class Tetromino extends Actor
{
    /**
     * Warna dari Tetromino.
     */
    protected Color color = Color.RED;
    /**
     * Variabel, untuk menampung blok blok.
     */
    private Block[] blocks = new Block[4];
    /**
     * Arah.
     */
    protected Direction dir = Direction.EAST;
    /**
     * Apakah tetromino ini (salah satu dari blok blok) sudah 'menapak'/menyentuh tanah.
     */
    private boolean isGround = false;
    /**
     * Apakah sudah tidak aktif?
     */
    private boolean isInactive = false;
    /**
     * Informasi posisi relatif blok-blok individual terhadap tetromino ini.
     */
    protected int[][] northBlocks = new int[4][2];
    protected int[][] eastBlocks = new int[4][2];
    protected int[][] southBlocks = new int[4][2];
    protected int[][] westBlocks = new int[4][2];

    /**
     * Melakukan "landslide".
     */
    boolean canSlide = true;

    /**
     * Jeda waktu sebelum turun.
     */
    private int w, speed = 10;

    /**
     * Enumerasi arah. UTARA, TIMUR, SELATAN, BARAT.
     */
    enum Direction {
        NORTH, EAST, SOUTH, WEST;

        /**
         * Memberikan arah selanjutnya dari arah yang dimasukan
         * kedalam parameter.
         * 
         * @param dir arah saat ini.
         * @return arah selanjutnya.
         */
        public static Direction next(Direction dir) {
            Direction[] dirs = Direction.values();
            int ord = dir.ordinal();
            return dirs[(ord+1)%dirs.length];
        }
        
        /**
         * Memberikan arah sebelumnya dari arah yang dimasukan.
         */
        public static Direction prev(Direction dir) {
            Direction[] dirs = Direction.values();
            int ord = dir.ordinal();

            return (ord-1<0)?dirs[Direction.values().length-1]:dirs[ord-1];
        }
    }

    /**
     * Pembentuk (Konstruktor) dari Tetromino.
     * 
     * @param color warna dari Tetromino, yang nantinya digunakan untuk mewarnai blok-blok.
     */
    public Tetromino(Color color) {
        //memasukan objek 'Blok' kedalam variabel. 
        for(int i=0;i<blocks.length;i++) {
            blocks[i] = new Block(this, color);
        }
        //Tetromino tidak terlihat, yang terlihat hanyalah kumpulan blok-bloknya saja.
        //Sehingga gambarnya hanya kecil (1x1 pixel berwarna putih)
        setImage(new GreenfootImage(1,1));
    }

    public Tetromino() {
        //memasukan objek 'Blok' kedalam variabel. 
        for(int i=0;i<blocks.length;i++) {
            blocks[i] = new Block(this, color);
        }
        //Tetromino tidak terlihat, yang terlihat hanyalah kumpulan blok-bloknya saja.
        //Sehingga gambarnya hanya kecil (1x1 pixel berwarna putih)
        setImage(new GreenfootImage(1,1));
    }

    /**
     * addedToWorld(World w) merupakan metode/fungsi buatan Greenfoot,
     * yang terlaksana apabila objek ini(Tetromino) masuk kedalam World, dalam game ini (TetrisWorld).
     * Kalimat yang pas buat fungsi ini adalah, 'Apa yang terjadi pada saat objek ini dimasukan kedalam dunia?'
     */
    public void addedToWorld(World w) {
        w.addObject(blocks[0], getX(), getY());
        w.addObject(blocks[1], getX(), getY());
        w.addObject(blocks[2], getX(), getY());
        w.addObject(blocks[3], getX(), getY());
        updateBlocksPosition();
    }

    /**
     * Gerak!
     */
    public void act() {
        /**
         * Kendali.. ini termasuk gameplaynya
         */
        String key = Greenfoot.getKey();
        if("right".equals(key)) { //apabila arah kanan ditekan
            moveRight(); //ke kanan
        }else if("left".equals(key)) { // apabila arah kiri ditekan
            moveLeft(); //ke kiri
        }
        else if("up".equals(key)) { // apabila tombol atas diputar
            if(canRotate()) {    // apabila bisa diputar
                nextDir(); // arah selanjutnya (diputar)
            }
        }
        
        /**
         * Ini mengatur gerak kebawahnya.
         */
        if(w > speed) { // jika w sudah sampai batasnya
            if(!isGround()) {  //jika tidak ditanah
                canSlide = true;
                setLocation(getX(), getY()+1); // turun satu
            }else if(isGround() && canSlide == true){ // jika tidak tidak ditanah (ditanah)
                canSlide = false; // tidak bisa slide.
            }else {
                inactive();
            }
            w=0; //w nya kembali ke 0
        }
        //posisi blok blok kita update sesuai dengan perubahan lokasi Tetromino.
        try {
            updateBlocksPosition();
        }catch(IllegalStateException e) {

        }
        w++; //w ditambah terus hingga w > speed
    }    

    /**
     * Pindah ke kanan.
     */
    private void moveRight() { 
        if(canRight() && !isGround()) { //jika bisa ke kanan DAN tidak di tanah
            setLocation(getX()+1, getY()); //belok kanan seperti biasa
            ((TetrisWorld)getWorld()).play(88, 100);
            ((TetrisWorld)getWorld()).play(90, 100);
            updateBlocksPosition();
        }else if(isGround() && canRight() && canSlide) { //untuk "landslide", apabila sudah ditanah, bisa ke kanan, dan bisa 'slide'
            setLocation(getX()+1, getY()); // belok kanan
            canSlide = false; // lalu canSlide nya di tiadakan
        }
    }
    
    /**
     * Pindah ke kiri.
     */
    private void moveLeft() {
        if(canLeft() && !isGround()){  // apabila bisa kekiri DAN tidak ditanah
            setLocation(getX()-1, getY()); // kekiri seperti biasa.
            updateBlocksPosition();
            ((TetrisWorld)getWorld()).play(88, 100);
            ((TetrisWorld)getWorld()).play(90, 100);
            
        }else if(isGround() && canLeft() && canSlide) { //apabila di tanah, bisa ke kiri, dan masih bisa "land slide"
            setLocation(getX()-1, getY()); // belok ke kiri
            canSlide = false; // landslidenya sudah tidak bisa.
        }
    }

    /**
     * Cek apakah bisa tetromino dari blok ini bisa berputar.
     * 
     * @return apakah posisi selanjutnya sah (tidak ada yang menghalangi ATAU diluar batas)
     */
    private boolean canRotate() {
        boolean result = isDirValid(getArray(Direction.next(dir)));
        return result;
    }

    /**
     * Apakah masing masing balok individual bisa ke kanan?
     * apa bila salah satu ada yang tidak bisa maka tidak bisa belok ke kanan.
     * 
     * @return  "bisa ke kanan?"
     */
    private boolean canRight() {
        if(isGround() && isInactive()) {return false;}
        int[][] dirArray = getArray(dir);
        boolean fb = isPosValid(dirArray[0][0]+1, dirArray[0][1]);
        boolean sb = isPosValid(dirArray[1][0]+1, dirArray[1][1]);
        boolean tb = isPosValid(dirArray[2][0]+1, dirArray[2][1]);
        boolean rb = isPosValid(dirArray[3][0]+1, dirArray[3][1]);
        return (fb&&sb&tb&&rb);
    }

    /**
     * Apakah masing masing balok individual bisa ke kiri?
     * apa bila salah satu ada yang tidak bisa maka tidak bisa belok ke kanan.
     * 
     * @return  "bisa ke kiri?"
     */
    private boolean canLeft() {
        if(isGround() && isInactive()) {return false;}
        int[][] dirArray = getArray(dir);
        boolean fb = isPosValid(dirArray[0][0]-1, dirArray[0][1]);
        boolean sb = isPosValid(dirArray[1][0]-1, dirArray[1][1]);
        boolean tb = isPosValid(dirArray[2][0]-1, dirArray[2][1]);
        boolean rb = isPosValid(dirArray[3][0]-1, dirArray[3][1]);
        return (fb&&sb&tb&&rb);
    }

    /**
     * Mengambil array posisi blok dari direksi.
     * 
     * @param direction direksi/arah yang ingin diketahui arraynya.
     * @return array dari direksi yang dimasukan.
     */
    private int[][] getArray(Direction direction) {
        switch(direction) {
            case NORTH: return northBlocks;
            case EAST:  return eastBlocks;
            case SOUTH: return southBlocks;
            case WEST:  return westBlocks;
            default: return null;
        }
    }

    /**
     * Cek apakah tetromino dari blok ini bisa pindah.
     * 
     * @param nextBlockPos informasi tentang direksi/arah berikutnya.
     * @return apakah keempat tempat untuk blok-blok tersedia.
     */
    private boolean isDirValid(int[][] nextBlockPos) {
        if(isGround()) {return false;}
        boolean fb = isPosValid(nextBlockPos[0][0], nextBlockPos[0][1]);
        boolean sb = isPosValid(nextBlockPos[1][0], nextBlockPos[1][1]);
        boolean tb = isPosValid(nextBlockPos[2][0], nextBlockPos[2][1]);
        boolean rb = isPosValid(nextBlockPos[3][0], nextBlockPos[3][1]);
        return (fb&&sb&tb&&rb);
    }

    /**
     * Mengecek apakah posisi x dan y yang kita masukan itu sah (tidak keluar ataupun bertemu dengan blok lain)
     * 
     * @param x posisi x relatif dari tetromino ini.
     * @param y posisi y relatif dari tetromino ini.
     * 
     * @return apakah posisi tersebut sah (tidak ada penghalang dan tidak diluar dunia).
     */
    private boolean isPosValid(int x, int y) {
        if(!(getX()+x >= 10 || getX()+x < 0)) { //apakah  posisi relatif tersebut tidak diluar dunia
            // cek apakah tidak ada blok yang menghalangi
            //try { 
            Block b = (Block) getOneObjectAtOffset(x,y, Block.class); // ambil blok dari posisi relatif tetromino
            if(b == null) { // jika tidak ada
                return  true; // sah
            }else if(b.getTetromino() == this) { // jika blok yang didapat tetrominonya sama dengan ini
                return true; // sah
            }else{ // jika tidak kosong (ada isinya/ada yang terambil)
                return false; // tidak sah (posisi tersebut tidak sah)
            }
            //}catch(NullPointerException e) {
            //    return true; // jaga-jaga, sebenarnya sudah ditangani oleh "if(b.equals(null)){...}", jika null/kosong, sah
            //}
        }
        return false; //tidak sah
    }

    /**
     * Apakah tetromino ini (kumpulan blok-blok individual) menyentuh tanah?
     */
    public boolean isGround() {
        //kalau tidak mengerti try & catch, skip dulu juga tidak apa apa. :D
        try {
            // blok ke satu atau kedua atau ketiga atau keempat ada yang menyentuh tanah?
            // bila ada salah satu blok ada yang menyentuh tanah, maka isGrounded-nya bernilai true, benar,
            // alias keseluruhan tetromino sudah 'napak'. 
            isGround = (blocks[0].isGrounded() || blocks[1].isGrounded() || blocks[2].isGrounded() || blocks[3].isGrounded());
        }catch(Exception e) {
            return false; //jaga-jaga juga ^_^
        }
        return isGround;
    }

    /**
     * Mengganti dengan direksi lainnya, Dengan kata lain memutar.
     */
    public void nextDir() {
        // jika ditanah, tidak bisa.
        if(isGround()){ return; } 
        // ubah direksi dengan selanjutnya.
        dir = Direction.next(dir);
        // update :D
        updateBlocksPosition();
        ((TetrisWorld)getWorld()).play(90, 100);
        ((TetrisWorld)getWorld()).play(92, 100);
        ((TetrisWorld)getWorld()).play(95, 100);
    }

    /**
     * Mengembalikan nilai arah.
     * 
     * @return arah sekarang.
     */
    public Direction getDirection() {
        return dir;
    }

    /**
     * Diupdate sesuai dengan informasi posisi relatif.
     */
    private void updateBlocksPosition() {
        int[][] nextArrayPos = getArray(dir);
        for(int i=0;i<nextArrayPos.length;i++) {
            blocks[i].setLocation(getX()+nextArrayPos[i][0], getY()+nextArrayPos[i][1]);
        }
    }     

    /**
     * Membuat tetromino ini tidak aktif.
     */
    public void inactive() {
        isInactive = true;
        for(Block b : blocks) {
            b.passive();
        }
        getWorld().removeObject(this);
    }

    /**
     * Apakah sudah tidak aktif?
     * 
     * @return apakah tetromino ini sudah tidak aktif.
     */
    public boolean isInactive(){
        return isInactive;
    }
}
